
from socketd import SocketD
from socketd.transport.core import Entity
from socketd.transport.core.Flags import Flags
from socketd.transport.core.Message import Message
from socketd.transport.core.entity.EntityDefault import EntityDefault
from socketd.transport.core.entity.MessageBuilder import MessageBuilder
from socketd.utils.StrUtils import StrUtils

from folkmq.client.MqMessage import MqMessage
from folkmq.common.MqConstants import MqConstants
from folkmq.common.MqMetasResolver import MqMetasResolver
from folkmq.common.MqMetasV1 import MqMetasV1
from folkmq.common.MqMetasV2 import MqMetasV2


class MqMetasResolverV1(MqMetasResolver):
    def version(self) -> int:
        return 1

    def get_sender(self, m: Entity) -> str:
        return m.meta_or_default(MqMetasV2.MQ_META_SENDER, "");

    def get_key(self, m: Entity) -> str:
        return m.meta_or_default(MqMetasV1.MQ_META_KEY, "");

    def get_tag(self, m: Entity) -> str:
        return m.meta_or_default(MqMetasV2.MQ_META_TAG, "");

    def get_topic(self, m: Entity) -> str:
        return m.meta_or_default(MqMetasV1.MQ_META_TOPIC, "");

    def get_consumer_group(self, m: Entity) -> str:
        return m.meta_or_default(MqMetasV1.MQ_META_CONSUMER_GROUP, "");

    def set_consumer_group(self, m: Entity, consumerGroup: str):
        m.put_meta(MqMetasV1.MQ_META_CONSUMER_GROUP, consumerGroup);

    def get_qos(self, m: Entity) -> int:
        return 0 if "0" == m.meta(MqMetasV1.MQ_META_QOS) else 1;

    def get_times(self, m: Entity) -> int:
        return int(m.meta_or_default(MqMetasV1.MQ_META_TIMES, "0"));

    def set_times(self, m: Entity, times: int):
        m.put_meta(MqMetasV1.MQ_META_TIMES, times.toString());

    def get_expiration(self, m: Entity) -> float:
        return float(m.meta_or_default(MqMetasV1.MQ_META_EXPIRATION, "0"));

    def set_expiration(self, m: Entity, expiration: int | None):
        if expiration is None:
            m.delMeta(MqMetasV1.MQ_META_EXPIRATION)
        else:
            m.put_meta(MqMetasV1.MQ_META_EXPIRATION, expiration.toString())

    def get_scheduled(self, m: Entity) -> int:
        return int(m.meta_or_default(MqMetasV1.MQ_META_SCHEDULED, "0"))

    def set_scheduled(self, m: Entity, scheduled: int):
        m.put_meta(MqMetasV1.MQ_META_SCHEDULED, scheduled.toString())

    def is_sequence(self, m: Entity) -> bool:
        return "1" == (m.meta_or_default(MqMetasV1.MQ_META_SEQUENCE, "0"))

    def is_broadcast(self, m: Entity) ->bool:
        return "1" == (m.meta(MqMetasV2.MQ_META_BROADCAST))

    def is_transaction(self, m: Entity) -> bool:
        return "1" == (m.meta(MqMetasV2.MQ_META_TRANSACTION))

    def set_transaction(self, m: Entity, isTransaction: bool):
        m.put_meta(MqMetasV2.MQ_META_TRANSACTION, ( "1" if isTransaction else "0"))

    def publish_entity_build(self, topic: str, message: MqMessage) -> EntityDefault:
        # 构建消息实体
        entity = EntityDefault().data_set(message.get_body())

        entity.meta_put(MqMetasV1.MQ_META_KEY, message.get_key())
        entity.meta_put(MqMetasV1.MQ_META_TOPIC, topic)
        entity.meta_put(MqMetasV1.MQ_META_QOS, ("0" if message.get_qos() == 0 else "1"))

        # 标签
        if message.get_tag():
            entity.meta_put(MqMetasV2.MQ_META_TAG, message.get_tag())


        # 定时派发
        if message.getScheduled():
            entity.meta_put(MqMetasV1.MQ_META_SCHEDULED, str(message.getScheduled().getTime()))
        else:
            entity.meta_put(MqMetasV1.MQ_META_SCHEDULED, "0")

        # 过期时间
        if message.get_expiration() is not None:
            entity.meta_put(MqMetasV1.MQ_META_EXPIRATION, str(message.get_expiration().getTime()))


        if  message.is_transaction():
            entity.meta_put(MqMetasV2.MQ_META_TRANSACTION, "1")


        if message.get_sender():
            entity.meta_put(MqMetasV2.MQ_META_SENDER, message.get_sender())

        if message.is_broadcast():
            entity.meta_put(MqMetasV2.MQ_META_BROADCAST, "1")

        # 是否有序
        if  message.is_sequence() or message.is_transaction():
            entity.at(MqConstants.BROKER_AT_SERVER_HASH)

        if message.is_sequence():
            entity.meta_put(MqMetasV1.MQ_META_SEQUENCE, "1")
        else:
            entity.at(MqConstants.BROKER_AT_SERVER)

        # 用户属性
        #message.getAttrMap().forEach((value, key, map) = > {
        #    entity.put_meta(MqConstants.MQ_ATTR_PREFIX + key, value)
        #})

        return entity

    def routing_message_build(self, topic: str, message: MqMessage) -> Message:
        entity = self.publish_entity_build(topic, message).at(MqConstants.BROKER_AT_SERVER)
        messageDefault =  MessageBuilder().flag(Flags.Message) \
                                .sid(StrUtils.guid()) \
                                .event(MqConstants.MQ_EVENT_PUBLISH) \
                                .entity(entity) \
                                .build()
        return messageDefault


class MqConstants:
    #版本号
    FOLKMQ_VERSION = "folkmq-version"

    #命名空间
    FOLKMQ_NAMESPACE = "folkmq-namespace"

    #元信息：消息主建
    MQ_META_KEY = "mq.tid"

    #元信息：消息主题
    MQ_META_TOPIC = "mq.topic"
    #元信息：消费者组
    MQ_META_CONSUMER_GROUP = "mq.consumer" # 此处不改动，算历史痕迹。保持向下兼容

    #元信息：消费回执
    MQ_META_ACK = "mq.ack"
    #元信息：执行确认
    MQ_META_CONFIRM = "mq.confirm"
    #元信息：批量处理
    MQ_META_BATCH = "mq.batch"
    #元信息：执行回滚
    MQ_META_ROLLBACK = "mq.rollback"
    #事件：订阅
    MQ_EVENT_SUBSCRIBE = "mq.event.subscribe"
    #事件：取消订阅
    MQ_EVENT_UNSUBSCRIBE = "mq.event.unsubscribe"
    #事件：发布
    MQ_EVENT_PUBLISH = "mq.event.publish"

    #事件：发布二次提交
    MQ_EVENT_PUBLISH2 = "mq.event.publish2"

    #事件：取消发布
    MQ_EVENT_UNPUBLISH = "mq.event.unpublish"
    #事件：派发
    MQ_EVENT_DISTRIBUTE = "mq.event.distribute"
    #事件：请求
    MQ_EVENT_REQUEST = "mq.event.request"
    #事件：保存快照
    MQ_EVENT_SAVE = "mq.event.save"

    #事件：加入集群
    MQ_EVENT_JOIN = "mq.event.join"

    #事件：接口
    MQ_API = "mq.api"

    #接口名
    API_NAME = "api.name"

    #接口访问令牌
    API_TOKEN = "api.token"

    #管理指令
    ADMIN_PREFIX = "admin."

    #管理视图 - 队列
    ADMIN_VIEW_QUEUE = "admin.view.queue"

    #管理队列 - 强制删除
    ADMIN_QUEUE_FORCE_DELETE = "admin.queue.force.delete"
    #管理队列 - 强制删除
    ADMIN_QUEUE_FORCE_CLEAR = "admin.queue.force.clear"
    #管理队列 - 强制派发
    ADMIN_QUEUE_FORCE_DISTRIBUTE = "admin.queue.force.distribute"

    #连接参数：ak
    PARAM_ACCESS_KEY = "ak"
    #连接参数: sk
    PARAM_ACCESS_SECRET_KEY = "sk"

    #主题与消息者间隔符
    SEPARATOR_TOPIC_CONSUMER_GROUP = "#"

    #经理人服务
    BROKER_AT_SERVER = "folkmq-server"

    #经理人服务
    BROKER_AT_SERVER_HASH = "folkmq-server!"

    #经理人所有服务
    BROKER_AT_SERVER_ALL = "folkmq-server*"

    #事件缓存队列消息费者
    MQ_TRAN_CONSUMER_GROUP = "!"

    #事件缓存队列消息费者
    MQ_ATTR_PREFIX = "!"

    #最大分片大小（1m）
    MAX_FRAGMENT_SIZE = 1024 * 1024


from typing import Callable, Dict

from socketd import SocketD
from socketd.cluster.ClusterClientSession import ClusterClientSession
from socketd.exception.SocketDExecption import SocketDConnectionException, SocketDException
from socketd.transport.client.ClientConfig import ClientConfig
from socketd.transport.core import Entity
from socketd.transport.core.EntityMetas import EntityMetas
from socketd.transport.core.Message import Message
from socketd.transport.core.Session import Session
from socketd.transport.core.Entity import Reply
from socketd.transport.core.entity.EntityDefault import EntityDefault
from socketd.transport.core.entity.StringEntity import StringEntity
from socketd.transport.stream.RequestStream import RequestStream
from socketd.utils.CompletableFuture import CompletableFuture
from socketd.utils.LogConfig import log
from socketd.utils.RunUtils import RunUtils

from folkmq.client.MqAlarm import MqAlarm
from folkmq.client.MqClient import MqClientInternal
from folkmq.client.MqClientListener import MqClientListener
from folkmq.client.MqMessage import MqMessage
from folkmq.client.MqMessageReceived import MqMessageReceived, MqMessageReceivedImpl
from folkmq.client.MqSubscription import MqSubscription
from folkmq.client.MqTransaction import MqTransaction, MqTransactionImpl
from folkmq.common.MqAssert import MqAssert
from folkmq.common.MqConstants import MqConstants
from folkmq.common.MqMetasV2 import MqMetasV2
from folkmq.common.MqTopicHelper import MqTopicHelper
from folkmq.common.MqUtils import MqUtils
from folkmq.exception.FolkmqException import FolkmqException


# 消息客户端默认实现
class MqClientDefault(MqClientInternal):
    def __init__(self, urls:tuple[str], clientListener:MqClientListener|None = None):
        self._urls:tuple[str] = urls

        if clientListener is not None:
            self._clientListener = clientListener
        else:
            self._clientListener =  MqClientListener()

        self._clientListener.init(self)

        self._autoAcknowledge = True

        self._name:str|None = None
        self._namespace:str|None = None
        self._clientSession:ClusterClientSession|None = None
        self._clientConfigHandler:Callable[[ClientConfig], None]|None = None
        self._subscriptionMap:Dict[str, MqSubscription] = {}

        self._transactionCheckback:Callable[[MqMessageReceived],None] = None
        self._listenHandler:Callable[[MqMessageReceived],None] = None



    def name(self) -> str:
        return self._name

    def name_as(self, name: str) -> 'MqClient':
        self._name = name
        return self

    def namespace(self) -> str:
        return self._namespace

    def namespace_as(self, namespace: str) -> 'MqClient':
        self._namespace = namespace
        return self

    async def connect(self) -> 'MqClient':
        serverUrls:[str] = []

        for url in self._urls:
            url = url.replace("folkmq:ws://", "sd:ws://")
            url = url.replace("folkmq:wss://", "sd:wss://")
            url = url.replace("folkmq://", "sd:tcp://")
            serverUrls.append(url)

        self._clientSession = await (SocketD.create_cluster_client(*serverUrls)
            .config(self.__config_handle)
            .listen(self._clientListener)
            .open())

        return self

    def __config_handle(self, c:ClientConfig):
        from folkmq.FolkMQ import FolkMQ

        (c.meta_put(MqConstants.FOLKMQ_VERSION, FolkMQ.version_code_as_string())
          .heartbeat_interval(6_000)
          .io_threads(1)
          .codec_threads(1)
          .exchange_threads(1))

        if self._name:
            c.meta_put("@", self._name)

        if self._namespace:
            c.meta_put(MqConstants.FOLKMQ_NAMESPACE, self._namespace)

        if self._clientConfigHandler:
            self._clientConfigHandler(c)

    async def disconnect(self):
        await self._clientSession.close()

    def config(self, configHandler: Callable[[ClientConfig], None]) -> 'MqClient':
        self._clientConfigHandler = configHandler
        return self

    def auto_acknowledge(self, auto: bool) -> 'MqClient':
        self._autoAcknowledge = auto
        return self

    async def subscribe(self, topic: str, consumerGroup: str | None, autoAck: bool | None,
                  consumerHandler: Callable[[MqMessageReceived], None]):
        if consumerGroup is None:
            consumerGroup = self.name()

        if autoAck is None:
            autoAck = self._autoAcknowledge

        MqAssert.require_non_null(topic, "Param 'topic' can't be null")
        MqAssert.require_non_null(consumerGroup, "Param 'consumerGroup' can't be null")
        MqAssert.require_non_null(consumerHandler, "Param 'consumerHandler' can't be null")

        MqAssert.assert_meta(topic, "topic")
        MqAssert.assert_meta(consumerGroup, "consumerGroup")

        #支持命名空间
        topic = MqTopicHelper.get_full_topic(self._namespace, topic)

        subscription = MqSubscription(topic, consumerGroup, autoAck, consumerHandler)
        self._subscriptionMap[subscription.get_queue_name()] = subscription

        if self._clientSession is not None:
            for session in self._clientSession.get_session_all():
                #如果有连接会话，则执行订阅
                entity = (StringEntity("")
                          .meta_put(MqConstants.MQ_META_TOPIC, subscription.get_topic())
                          .meta_put(MqConstants.MQ_META_CONSUMER_GROUP, subscription.get_consumer_group())
                          .meta_put(EntityMetas.META_X_UNLIMITED, "1")
                          .at(MqConstants.BROKER_AT_SERVER_ALL))
                #使用 Qos1
                await session.send_and_request(MqConstants.MQ_EVENT_SUBSCRIBE, entity, 30_000).waiter()
                log.info(f"Client subscribe successfully: {topic}#{consumerGroup}, sessionId={session.session_id()}")

    async def unsubscribe(self, topic: str, consumerGroup: str | None):
        if consumerGroup is None:
            consumerGroup = self.name()

        MqAssert.require_non_null(topic, "Param 'topic' can't be null")
        MqAssert.require_non_null(consumerGroup, "Param 'consumerGroup' can't be null")

        MqAssert.assert_meta(topic, "topic")
        MqAssert.assert_meta(consumerGroup, "consumerGroup")

        #支持命名空间
        topic = MqTopicHelper.get_full_topic(self._namespace, topic)

        queueName = topic + MqConstants.SEPARATOR_TOPIC_CONSUMER_GROUP + consumerGroup
        self._subscriptionMap.pop(queueName)

        if self._clientSession is not None:
            for session in self._clientSession.get_session_all():
                #如果有连接会话，则执行订阅
                entity = (StringEntity("")
                          .meta_put(MqConstants.MQ_META_TOPIC, topic)
                          .meta_put(MqConstants.MQ_META_CONSUMER_GROUP, consumerGroup)
                          .at(MqConstants.BROKER_AT_SERVER_ALL))
                #使用 Qos1
                await session.send_and_request(MqConstants.MQ_EVENT_UNSUBSCRIBE, entity, 30_000).waiter()
                log.info(f"Client unsubscribe successfully: {topic}#{consumerGroup}， sessionId={session.session_id()}")

    async def publish(self, topic: str, message: MqMessage):
        MqAssert.require_non_null(topic, "Param 'topic' can't be null")
        MqAssert.require_non_null(message, "Param 'message' can't be null")

        MqAssert.assert_meta(topic, "topic")

        if self._clientSession is None:
            raise SocketDConnectionException("Not connected!")

        # 支持命名空间
        topic = MqTopicHelper.get_full_topic(self._namespace, topic)
        session = self._clientSession.get_session_any(self._diversionOrNull(topic, message))

        if session is None or session.is_valid() == False:
            raise SocketDException("No session is available!")

        entity = MqUtils.get_of(session).publish_entity_build(topic, message)

        if message.get_qos() > 0:
            resp = await session.send_and_request(MqConstants.MQ_EVENT_PUBLISH, entity, 0).waiter()
            confirm = int(resp.meta_or_default(MqConstants.MQ_META_CONFIRM, "0"))
            if confirm != 1:
                messsage = "Client message publish confirm failed: " + resp.data_as_string()
                raise FolkmqException(messsage)
        else:
            session.send(MqConstants.MQ_EVENT_PUBLISH, entity)

    def publish_async(self, topic: str, message: MqMessage) -> CompletableFuture:
        MqAssert.require_non_null(topic, "Param 'topic' can't be null")
        MqAssert.require_non_null(message, "Param 'message' can't be null")

        MqAssert.assert_meta(topic, "topic")

        if self._clientSession is None:
            raise SocketDConnectionException("Not connected!")

        # 支持命名空间
        topic = MqTopicHelper.get_full_topic(self._namespace, topic)
        session = self._clientSession.get_session_any(self._diversionOrNull(topic, message))

        if session is None or session.is_valid() == False:
            raise SocketDException("No session is available!")

        entity = MqUtils.get_of(session).publish_entity_build(topic, message)

        if message.get_qos() > 0:
            _future :CompletableFuture  = CompletableFuture()
            def _then_reply_do(resp:Reply):
                confirm = int(resp.meta_or_default(MqConstants.MQ_META_CONFIRM, "0"))
                if confirm != 1:
                    messsage = "Client message publish confirm failed: " + resp.data_as_string()
                    _future.set_e(FolkmqException(messsage))
                    _future.accept(False)
                else:
                    _future.accept(True)

            def _then_error_do(err:Exception):
                _future.set_e(err)
                _future.accept(False)

            (session.send_and_request(MqConstants.MQ_EVENT_PUBLISH, entity, 0)
             .then_reply(_then_reply_do)
             .then_error(_then_error_do))

            return _future

        else:
            session.send(MqConstants.MQ_EVENT_PUBLISH, entity)
            return None

    async def unpublish(self, topic: str, key: str):
        MqAssert.require_non_null(topic, "Param 'topic' can't be null")
        MqAssert.require_non_null(key, "Param 'key' can't be null")

        MqAssert.assert_meta(topic, "topic")
        MqAssert.assert_meta(key, "key")

        if self._clientSession is None:
            raise SocketDConnectionException("Not connected!")

        #支持命名空间
        topic = MqTopicHelper.get_full_topic(self._namespace, topic)
        session = self._clientSession.get_session_any(None)
        if session is None or session.is_valid() == False:
            raise SocketDException("No session is available!")

        entity = (StringEntity("")
                  .meta_put(MqConstants.MQ_META_TOPIC, topic)
                  .meta_put(MqConstants.MQ_META_KEY, key)
                  .at(MqConstants.BROKER_AT_SERVER_ALL))

        resp = await session.send_and_request(MqConstants.MQ_EVENT_UNPUBLISH, entity).waiter()
        confirm = int(resp.meta_or_default(MqConstants.MQ_META_CONFIRM, "0"))
        if confirm != 1:
            messsage = "Client message unpublish confirm failed: " + resp.data_as_string()
            raise FolkmqException(messsage)



    async def listen(self, listenHandler: Callable[[MqMessageReceived], None]):
        # 检查必要条件
        if self._name is None:
            return FolkmqException("Client 'name' can't be empty")

        self._listenHandler = listenHandler

    def send(self, message: MqMessage, toName: str, timeout: int | None = None) -> RequestStream | None:
        # 检查必要条件
        if self._name is None:
            return FolkmqException("Client 'name' can't be empty")

        # 检查参数
        MqAssert.require_non_null(toName, "Param 'toName' can't be null")
        MqAssert.require_non_null(message, "Param 'message' can't be null")

        MqAssert.assert_meta(toName, "toName")

        if self._clientSession is None:
            raise SocketDConnectionException("Not connected!")

        session = self._clientSession.get_session_any(None)
        if session is None or session.is_valid() == False:
            raise SocketDException("No session is available!")

        message.internal_sender(self.name())
        entity = MqUtils.get_of(session).publish_entity_build("", message)
        entity.put_meta(MqMetasV2.MQ_META_CONSUMER_GROUP, toName)
        entity.at(toName)

        if message.get_qos() > 0:
            return session.send_and_request(MqConstants.MQ_EVENT_REQUEST, entity, timeout)
        else:
            session.send(MqConstants.MQ_EVENT_REQUEST, entity)
            return None

    def transaction_checkback(self, transactionCheckback: Callable[[MqMessageReceived], None]):
        if transactionCheckback is not None:
            self._transactionCheckback = transactionCheckback
        return self

    def new_transaction(self) -> MqTransaction:
        if self._name is None:
            return FolkmqException("Client 'name' can't be empty")

        return MqTransactionImpl(self)

    async def publish2(self, tmid: str, keyAry: [str], isRollback: bool):
        if keyAry is None or len(keyAry) == 0:
            return

        if self._clientSession is None:
            raise SocketDConnectionException("Not connected!")

        session = self._clientSession.get_session_any(tmid)

        if session is None or session.is_valid() == False:
            raise SocketDException("No session is available!")

        entity = (StringEntity(",".join(keyAry))
                  .meta_put(MqConstants.MQ_META_ROLLBACK, ( "1" if isRollback else "0"))
                  .at(MqConstants.BROKER_AT_SERVER_HASH)) # 事务走哈希

        resp = await session.send_and_request(MqConstants.MQ_EVENT_PUBLISH2, entity).waiter()

        confirm = int(resp.meta_or_default(MqConstants.MQ_META_CONFIRM, "0"))
        if confirm != 1:
            messsage = "Client message publish2 confirm failed: " + resp.data_as_string()
            raise FolkmqException(messsage)


    def reply(self, session: Session, f: Message, message: MqMessageReceivedImpl, isOk: bool, entity: Entity):
        """发送“回执”，向服务端反馈消费情况"""
        if message.get_qos() > 0:
            if session.is_valid():
                if entity is None:
                    entity = EntityDefault()

                if isinstance(entity, MqAlarm):
                    RunUtils.taskTry(session.send_alarm(f, entity.data_as_string()))
                else:
                    entity.put_meta(MqConstants.MQ_META_ACK, "1" if isOk else "0")
                    RunUtils.taskTry(session.reply_end(f, entity))

    def _diversionOrNull(self, fullTopic: str, message: MqMessage) -> str | None:
        if message.is_transaction():
            return message.get_tmid()
        elif message.is_sequence():
            if message.getSequenceSharding():
                return message.getSequenceSharding()
            else:
                return fullTopic
        else:
            return None

    def getSubscription(self, fullTopic: str, consumerGroup: str):
        queueName = fullTopic + MqConstants.SEPARATOR_TOPIC_CONSUMER_GROUP + consumerGroup
        return self._subscriptionMap.get(queueName)

    def getSubscriptionAll(self) -> [MqSubscription]:
        return self._subscriptionMap.values()
    def getSubscriptionSize(self):
        return len(self._subscriptionMap)


import json
import traceback

from socketd.exception.SocketDExecption import SocketDAlarmException
from socketd.transport.core.EntityMetas import EntityMetas
from socketd.transport.core.Message import Message
from socketd.transport.core.Session import Session
from socketd.transport.core.entity.StringEntity import StringEntity
from socketd.transport.core.listener.EventListener import EventListener
from socketd.utils.LogConfig import log

from folkmq.client.MqMessageReceived import MqMessageReceivedImpl
from folkmq.common.MqConstants import MqConstants


class MqClientListener(EventListener):
    def init(self, client: 'MqClientDefault'):
        """初始化"""
        self._client = client

        self.do_on(MqConstants.MQ_EVENT_DISTRIBUTE, self.doOn_distribute)
        self.do_on(MqConstants.MQ_EVENT_REQUEST, self.doOn_request)

    def doOn_distribute(self, s:Session, m:Message):
        try:
            message = MqMessageReceivedImpl(self._client, s, m)
            self.onReceive(s,m,message,False)
        except Exception as e:
            e_msg = traceback.format_exc()
            log.warning(f"Client consume handle error, sid={m.sid()} \n{e_msg}")

    def doOn_request(self, s:Session, m:Message):
        try:
            message = MqMessageReceivedImpl(self._client, s, m)
            self.onReceive(s,m,message,True)
        except Exception as e:
            e_msg = traceback.format_exc()
            log.warning(f"Client consume handle error, sid={m.sid()} \n{e_msg}")

    def onReceive(self, s: Session, m: Message, message: MqMessageReceivedImpl, isRequest: bool):
        """接收时"""
        if isRequest:
            try:
                if message.is_transaction():
                    if self._client._transactionCheckback is not None:
                        self._client._transactionCheckback(message)
                    else:
                        s.send_alarm(m, "Client no checkback handler!")
                else:
                    if self._client._listenHandler is not None:
                        self._client._listenHandler(message)
                    else:
                        s.send_alarm(m, "Client no request handler!")
            except Exception as e:
                try:
                    e_msg = traceback.format_exc()
                    if s.is_valid():
                        s.send_alarm(m, "Client request handle error:" + e)
                    log.warning(f"Client request handle error, key={message.get_key()} \n{e_msg}")
                except Exception as err:
                    err_msg = traceback.format_exc()
                    log.warning(f"Client request handle error, key={message.get_key()} \n{err_msg}")
        else:
            subscription = self._client.getSubscription(message.get_full_topic(), message.get_consumer_group())

            try:
                if subscription is not None:
                    # 有订阅
                    subscription.consume(message)
                    if subscription.is_auto_ack():
                        self._client.reply(s, m, message, True, None)
                else:
                    # 没有订阅
                    self._client.reply(s, m, message, False, None)
            except Exception as e:
                try:
                    if subscription is not None:
                        # 有订阅
                        if subscription.is_auto_ack():
                            self._client.reply(s, m, message, False, None)
                    else:
                        # 没有订阅
                        self._client.reply(s, m, message, False, None)

                    e_msg = traceback.format_exc()
                    log.warning(f"Client consume handle error, key={message.get_key()} \n{e_msg}")
                except Exception as err:
                    err_msg = traceback.format_exc()
                    log.warning(f"Client consume handle error, key={message.get_key()} \n{err_msg}")

    async def on_open(self, session:Session):
        """会话打开时"""
        await super().on_open(session)

        log.info(f"Client session opened, sessionId={session.session_id()}")

        if self._client.getSubscriptionSize() == 0:
            return

        subscribeData:dict[str,str] = {}
        for subscription in self._client.getSubscriptionAll():
            queueNameSet = subscribeData.get(subscription.get_topic())
            if queueNameSet is None:
                queueNameSet = set()
                subscribeData[subscription.get_topic()] = queueNameSet

            queueNameSet.add(subscription.get_queue_name())

        jsonStr = json.dumps(subscribeData)
        entity = (StringEntity(jsonStr)
                  .meta_put(MqConstants.MQ_META_BATCH, "1")
                  .meta_put(EntityMetas.META_X_UNLIMITED, "1")
                  .meta_put("@", MqConstants.BROKER_AT_SERVER))

        await session.send_and_request(MqConstants.MQ_EVENT_SUBSCRIBE, entity).waiter()

        log.info(f"Client onOpen batch subscribe successfully, sessionId={session.session_id()}")

    def on_close(self, session: Session):
        """会话关闭时"""
        super().on_close(session)

        log.info(f"Client session closed, sessionId={session.session_id()}")

    async def on_error(self, session: Session, error: Exception):
        """会话出错时"""
        await super().on_error(session, error)

        if isinstance(error, SocketDAlarmException):
            log.warning(f"Client alarm, sessionId={session.session_id()}, alarm:{error}")
        else:
            log.warning(f"Client error, sessionId={session.session_id()}, error:{error}")



# 消息元信息 v2
#
# @author noear
# @since 1.1
class MqMetasV2:
    #
    # 元信息：消息主建
    #
    MQ_META_KEY = "t0"
    #
    # 元信息：消息元信息版本id
    #
    MQ_META_VID = "v0"

    #
    # 元信息：发送人
    #
    MQ_META_SENDER = "s0"

    #
    # 元信息：消息主题
    #
    MQ_META_TOPIC = "t1"

    #
    # 元信息：消费者组
    #
    MQ_META_CONSUMER_GROUP = "c1"  # 此处不改动，算历史痕迹。保持向下兼容

    #
    # 元信息：消息调度时间
    #
    MQ_META_SCHEDULED = "s1"
    #
    # 元信息：消息过期时间
    #
    MQ_META_EXPIRATION = "e1"
    #
    # 元信息：消息是否有序
    #
    MQ_META_SEQUENCE = "s2"
    #
    # 元信息：消息是否事务
    #
    MQ_META_TRANSACTION = "t4"

    #
    # 元信息：消息是否广播
    #
    MQ_META_BROADCAST = "b0"

    #
    # 元信息：消息质量等级
    #
    MQ_META_QOS = "q1"
    #
    # 元信息：派发次数
    #
    MQ_META_TIMES = "t2"
    #
    # 元信息：标签
    #
    MQ_META_TAG = "t5"


from typing import Callable

from folkmq.client.MqMessageReceived import MqMessageReceived
from folkmq.common.MqConstants import MqConstants


class MqSubscription:
    def __init__(self, topic: str, consumerGroup: str, autoAck: bool, consumeHandler:Callable[[MqMessageReceived], None]):
        self._topic = topic
        self._consumerGroup = consumerGroup
        self._autoAck = autoAck
        self._queueName = topic + MqConstants.SEPARATOR_TOPIC_CONSUMER_GROUP + consumerGroup
        self._consumeHandler = consumeHandler

    def get_topic(self) -> str:
        return self._topic

    def get_consumer_group(self) -> str:
        return self._consumerGroup

    def is_auto_ack(self) -> bool:
        return self._autoAck

    def get_queue_name(self) -> str:
        return self._queueName

    def consume(self, message:MqMessageReceived):
        self._consumeHandler(message)


# 消息元信息 v1
#
# @author noear
# @since 1.1
class MqMetasV1:
    #
    # 元信息：消息主建
    #
    MQ_META_KEY = "mq.tid"

    #
    # 元信息：消息主题
    #
    MQ_META_TOPIC = "mq.topic"
    #
    # 元信息：消费者组
    #
    MQ_META_CONSUMER_GROUP = "mq.consumer"  # 此处不改动，算历史痕迹。保持向下兼容

    #
    # 元信息：消息调度时间
    #
    MQ_META_SCHEDULED = "mq.scheduled"
    #
    # 元信息：消息过期时间
    #
    MQ_META_EXPIRATION = "mq.expiration"
    #
    # 元信息：消息是否有序
    #
    MQ_META_SEQUENCE = "mq.sequence"

    #
    # 元信息：消息质量等级
    #
    MQ_META_QOS = "mq.qos"
    #
    # 元信息：派发次数
    #
    MQ_META_TIMES = "mq.times"


from typing import Union, Callable, Dict

from folkmq.client.MqMessageReceived import MqMessageReceived


# 消息路由器
class MqRouter:
    def __init__(self, mappingHandler: Callable[[MqMessageReceived], str]):
        self._mappingHandler:Callable[[MqMessageReceived], str] = mappingHandler
        self._mappingMap:Dict[str,str] = {}
        self._consumeHandler:Union[Callable[[MqMessageReceived], None], None] = None

    # 添加映射处理
    def do_on(self, mapping:str, consumeHandler:Callable[[MqMessageReceived], None]) -> 'MqRouter':
        self._mappingMap[mapping] = consumeHandler
        return self

    # 添加消费处理
    def do_on_consume(self, consumeHandler:Callable[[MqMessageReceived], None]) -> 'MqRouter':
        self._consumeHandler = consumeHandler
        return self

    # 消费
    def consume(self, message:MqMessageReceived):
        if self._consumeHandler:
            self._consumeHandler(message)

        mapping = self._mappingHandler(message)
        handler = self._mappingMap.get(mapping)
        if handler:
            handler(message)


from socketd.transport.core.Message import Message
from socketd.transport.core.Session import Session
from socketd.transport.core.entity.MessageDefault import MessageDefault

from folkmq.common.MqConstants import MqConstants
from folkmq.common.MqMetasResolver import MqMetasResolver
from folkmq.common.MqMetasResolverV1 import MqMetasResolverV1
from folkmq.common.MqMetasResolverV2 import MqMetasResolverV2
from folkmq.common.MqMetasV2 import MqMetasV2

# 消息工具类
class MqUtils:
    __v1 = MqMetasResolverV1()
    __v2 = MqMetasResolverV2()

    @staticmethod
    def get_of(t: Session | Message | None) -> MqMetasResolver:
        if t is None:
            return MqUtils.__v2
        else:
            if isinstance(t, MessageDefault):
                ver = t.meta_or_default(MqMetasV2.MQ_META_VID, "1")
                if "1" == ver:
                    return MqUtils.__v1
                else:
                    return MqUtils.__v2
            else:
                ver = t.handshake().param_or_default(MqConstants.FOLKMQ_VERSION, "1")
                if "1" == ver:
                    return MqUtils.__v1
                else:
                    return MqUtils.__v2



class MqAssert:
    @staticmethod
    def require_non_null(obj, message:str):
        if obj is None:
            raise Exception(message)
        else:
            return obj
    @staticmethod
    def assert_meta(str:str, paramName:str):
        MqAssert.assert_meta_symbols(str, paramName, '?', "?")
        MqAssert.assert_meta_symbols(str, paramName, '&', "&")
        MqAssert.assert_meta_symbols(str, paramName, '=', "=")
        MqAssert.assert_meta_symbols(str, paramName, '!', "!")

        MqAssert.assert_meta_symbols(str, paramName, '@', "@")
        MqAssert.assert_meta_symbols(str, paramName, '#', "#")
        MqAssert.assert_meta_symbols(str, paramName, '$', "$")
        MqAssert.assert_meta_symbols(str, paramName, '%', "%")
        MqAssert.assert_meta_symbols(str, paramName, '^', "^")
        MqAssert.assert_meta_symbols(str, paramName, '*', "*")

    @staticmethod
    def assert_meta_symbols(str: str, paramName: str, c:str, cS:str):
        if str.find(c) >= 0:
            raise Exception("Param '" + paramName + "' can't have symbols: '" + cS + "'");



from abc import abstractmethod, ABC
from asyncio import Future
from typing import Callable

from socketd.transport.client.ClientConfig import ClientConfig
from socketd.transport.core import Entity
from socketd.transport.core.Message import Message
from socketd.transport.core.Session import Session
from socketd.transport.stream.RequestStream import RequestStream
from socketd.utils.CompletableFuture import CompletableFuture

from folkmq.client.MqMessage import MqMessage
from folkmq.client.MqMessageReceived import MqMessageReceived, MqMessageReceivedImpl
from folkmq.client.MqTransaction import MqTransaction

# 消息客户端
class MqClient(ABC):
    @abstractmethod
    def name(self) -> str:
        """名字（即，默认消费者组）"""
        ...

    @abstractmethod
    def name_as(self, name: str) -> 'MqClient':
        """名字取为（即，默认消费者组）"""
        ...

    @abstractmethod
    def namespace(self) -> str:
        """命名空间"""
        ...

    @abstractmethod
    def namespace_as(self, namespace: str) -> 'MqClient':
        """命名空间"""
        ...

    @abstractmethod
    async def connect(self) -> 'MqClient':
        """连接"""
        ...

    @abstractmethod
    async def disconnect(self):
        """断开连接"""
        ...

    @abstractmethod
    def config(self, configHandler: Callable[[ClientConfig], None]) -> 'MqClient':
        """客户端配置"""
        ...

    @abstractmethod
    def auto_acknowledge(self, auto: bool) -> 'MqClient':
        """
        自动回执
        :param auto: 自动（默认为 true）
        """
        ...

    @abstractmethod
    async def subscribe(self, topic: str, consumerGroup: str | None, autoAck: bool | None,
                  consumerHandler: Callable[[MqMessageReceived], None]):
        """
        订阅主题
        :param topic:           主题
        :param consumerGroup:   消费者组
        :param consumerHandler: 消费处理
        """
        ...

    @abstractmethod
    async def unsubscribe(self, topic: str, consumerGroup: str | None):
        """
        取消订阅主题
        param topic:         主题
        param consumerGroup: 消费者组
        """
        ...

    @abstractmethod
    async def publish(self, topic: str, message: MqMessage):
        """
        同步发布消息
        :param topic:   主题
        :param message: 消息
        """
        ...

    def publish_async(self, topic: str, message: MqMessage) -> CompletableFuture:
        """
        异步发布消息
        :param topic:   主题
        :param message: 消息
        """
        ...

    @abstractmethod
    async def unpublish(self, topic: str, key: str):
        """
        取消发布
        :param topic: 主题
        :param key:   主建
        """
        ...


    @abstractmethod
    async def listen(self, listenHandler: Callable[[MqMessageReceived], None]):
        """
        监听
        :param listenHandler: 监听处理
        :return:
        """
        ...


    @abstractmethod
    def send(self, message: MqMessage, toName: str, timeout: int | None) -> RequestStream | None:
        """发送
        :param message:消息
        :param toName:发送目标名字
        :param timeout:超时（单位毫秒）
        """
        ...

    @abstractmethod
    def transaction_checkback(self, transactionCheckback: Callable[[MqMessageReceived], None]) -> 'MqClient':
        """事务回查
        :param transactionCheckback: 事务回查处理
        """
        ...


    @abstractmethod
    def new_transaction(self) -> MqTransaction:
        """新建事务"""
        ...


class MqClientInternal(MqClient):
    @abstractmethod
    async def publish2(self, tmid: str, keyAry: [str], isRollback: bool):
        """发布二次提交"""
        ...

    @abstractmethod
    def reply(self, session: Session, f: Message, message: MqMessageReceivedImpl, isOk: bool, entity: Entity):
        """消费答复"""
        ...


from abc import abstractmethod
from datetime import datetime

from socketd.transport.core import Entity
from socketd.transport.core.Message import Message
from socketd.transport.core.Session import Session

from folkmq.client.MqMessage import MqMessageBase
from folkmq.common.MqConstants import MqConstants
from folkmq.common.MqTopicHelper import MqTopicHelper
from folkmq.common.MqUtils import MqUtils


class MqMessageReceived(MqMessageBase):
    # 主题
    @abstractmethod
    def get_topic(self) -> str:
        ...

    # 内容
    @abstractmethod
    def get_body_as_string(self) -> str:
        ...

    # 消费者组
    @abstractmethod
    def get_consumer_group(self) -> str:
        ...

    # 已派发次数
    @abstractmethod
    def get_times(self) -> int:
        ...

    # 回执
    @abstractmethod
    def acknowledge(self, isOk: bool):
        ...

    # 响应
    @abstractmethod
    def response(self, entity:Entity):
        ...

class MqMessageReceivedImpl(MqMessageReceived):
    def __init__(self, clientInternal:'MqClientInternal', session:Session, source:Message):
        self.__clientInternal = clientInternal
        self.__session = session
        self.__source = source

        mr  = MqUtils.get_of(source)

        self.__sender = mr.get_sender(source)

        self.__key = mr.get_key(source)
        self.__tag = mr.get_tag(source)
        self.__fullTopic = mr.get_topic(source)
        self.__topic = MqTopicHelper.get_topic(self.__fullTopic)
        self.__consumerGroup = mr.get_consumer_group(source)

        self.__qos = mr.get_qos(source)
        self.__times = mr.get_times(source)
        self.__sequence = mr.is_sequence(source)
        self.__broadcast = mr.is_broadcast(source)
        self.__transaction = mr.is_transaction(source)

        self.__expirationL = mr.get_expiration(source)
        self.__expiration = self.__expirationL

    def getSource(self) -> Message:
        return self.__source

    def get_sender(self) -> str | None:
        return self.__sender

    def get_key(self) -> str:
        return self.__key

    def get_tag(self) -> str:
        return self.__tag

    def get_topic(self) -> str:
        return self.__topic

    def get_full_topic(self)->str:
        return self.__fullTopic

    def get_consumer_group(self) -> str:
        return self.__consumerGroup

    def get_body(self) -> bytes:
        return self.__source.data_as_bytes()

    def get_body_as_string(self) -> str:
        return self.__source.data_as_string()

    def get_qos(self) -> int:
        return self.__qos

    def get_attr(self, name: str) -> str | None:
        return self.__source.meta(MqConstants.MQ_ATTR_PREFIX + name)

    def get_expiration(self) -> datetime | None:
        return self.__expiration

    def is_transaction(self) -> bool:
        return self.__transaction

    def is_sequence(self) -> bool:
        return self.__sequence

    def is_broadcast(self) -> bool:
        return self.__broadcast

    def get_times(self) -> int:
        return self.__times

    def acknowledge(self, isOk: bool):
        self.__clientInternal.reply(self.__session, self.__source, self, isOk, None)

    def response(self, entity:Entity):
        self.__clientInternal.reply(self.__session, self.__source, self, True, entity)

    def __str__(self) ->str:
        return "MqMessageReceived{" + \
            "key='" + self.__key + '\'' + \
            ", tag='" + self.__tag + '\'' + \
            ", topic='" + self.__topic + '\'' + \
            ", body='" + self.get_body_as_string() + '\'' + \
            '}'





from abc import abstractmethod, ABC

from socketd.utils.StrUtils import StrUtils


class MqTransaction(ABC):
    @abstractmethod
    def tmid(self) -> str:
        """事务管理id"""
        ...

    @abstractmethod
    def binding(self, message: 'MqMessage'):
        """事务绑定"""
        ...

    @abstractmethod
    async def commit(self):
        """事务提交"""
        ...

    @abstractmethod
    async def rollback(self):
        """事务回滚"""
        ...


class MqTransactionImpl(MqTransaction):
    def __init__(self, client: 'MqClientInternal'):
        self._client = client
        self._keyAry:[str] = []
        self._tmid:str = StrUtils.guid()

    def tmid(self) -> str:
        return self._tmid

    def binding(self, message: 'MqMessage'):
        self._keyAry.append(message.get_key())
        message.internal_sender(self._client.name())

    async def commit(self):
        await self._client.publish2(self._tmid, self._keyAry, False)

    async def rollback(self):
        await self._client.publish2(self._tmid, self._keyAry, True)


from abc import abstractmethod
from datetime import datetime

from socketd.utils.StrUtils import StrUtils

from folkmq.client.MqTransaction import MqTransaction


class MqMessageBase:
    #发送人
    @abstractmethod
    def get_sender(self) -> str | None: ...

    #主建
    @abstractmethod
    def get_key(self) -> str: ...

    #标签
    @abstractmethod
    def get_tag(self) -> str: ...

    #内容
    @abstractmethod
    def get_body(self) -> bytes: ...

    #过期时间
    @abstractmethod
    def get_expiration(self) -> datetime | None: ...

    #是否事务
    @abstractmethod
    def is_transaction(self) -> bool: ...

    #是否有序
    @abstractmethod
    def is_sequence(self) -> bool: ...

    # 是否广播
    @abstractmethod
    def is_broadcast(self) -> bool: ...

    #质量等级
    @abstractmethod
    def get_qos(self) -> int: ...

    #获取属性
    @abstractmethod
    def get_attr(self, name: str) -> str | None: ...


class MqMessage(MqMessageBase):
    def __init__(self, body: str|bytes, key:str|None = None) :
        if key:
            self.__key = key
        else:
            self.__key = StrUtils.guid()

        if isinstance(body, str):
            self.__body = body.encode("utf-8")
        else:
            self.__body = body

        self.__sender = None;
        self.__tag = None;
        self.__scheduled:datetime = None
        self.__expiration:datetime = None;
        self.__sequence:bool = False
        self.__broadcast:bool = False
        self.__sequenceSharding: str | None = None
        self.__qos:int = 1

        self.__attrMap: dict[str, str] = {}
        self.__transaction:MqTransaction = None

    def get_sender(self) -> str | None:
        return self.__sender

    def get_key(self) -> str:
        return self.__key

    def get_tag(self) -> str | None:
        return self.__tag

    def get_body(self) -> bytes:
        return self.__body

    def getScheduled(self) -> datetime:
        return self.__scheduled

    def get_expiration(self) -> datetime:
        return self.__expiration

    def is_transaction(self) -> bool:
        return self.__transaction is not None

    def is_broadcast(self) -> bool:
        return self.__broadcast

    def is_sequence(self) -> bool:
        return self.__sequence

    def getSequenceSharding(self)->str:
        return self.__sequenceSharding

    def get_qos(self) -> int:
        return self.__qos

    def tag(self, tag: str)->'MqMessage':
        self.__tag = tag
        return self

    def as_json(self)-> 'MqMessage':
        self.attr("Content-Type", "application/json")
        return self;



    def scheduled(self, scheduled: datetime)-> 'MqMessage':
        self.__scheduled = scheduled
        return self


    def expiration(self,expiration: datetime)-> 'MqMessage':
        self.__expiration = expiration
        return self

    def sequence(self, sequence: bool, sharding: str|None)-> 'MqMessage':
        self.__sequence = sequence
        if sequence:
            if StrUtils.is_not_empty(sharding):
                self.__sequenceSharding = sharding
        else:
            self.__sequenceSharding = None
        return self

    def broadcast(self, broadcast: bool)-> 'MqMessage':
        self.__broadcast = broadcast
        return self


    def transaction(self, transaction: MqTransaction|None)-> 'MqMessage':
        if transaction is not None:
            self.__transaction = transaction
            transaction.binding(self)

        return self

    def get_tmid(self)-> str | None:
        if self.__transaction is None:
            return None
        else:
            return self.__transaction.tmid()

    def internal_sender(self, sender: str)-> 'MqMessage':
        self.__sender = sender
        return self

    def qos(self, qos: int)-> 'MqMessage':
        self.__qos = qos
        return self

    def get_attr(self, name: str)-> str | None:
        tmp = self.__attrMap.get(name);
        return tmp;


    def get_attr_map(self)-> dict[str,str]:
        return self.__attrMap


    def attr(self, name: str, value: str)-> 'MqMessage':
        self.__attrMap.set(name, value)
        return self






class MqTopicHelper:
    #获取完整主题
    @staticmethod
    def get_full_topic(namespace:str, topic:str):
        if namespace:
            return f"{namespace}:{topic}"
        else:
            return topic

    #获取主题
    @staticmethod
    def get_topic(fullTopic:str):
        idx = fullTopic.find(":")
        if  idx > 0:
            return fullTopic[idx + 1]
        else:
            return fullTopic


class MqApis:
    MQ_QUEUE_LIST = "mq.queue.list"
    MQ_QUEUE_VIEW_MESSAGE = "mq.queue.view.message"
    MQ_QUEUE_VIEW_SESSION = "mq.queue.view.session"
    MQ_QUEUE_FORCE_CLEAR = "mq.queue.force.clear"
    MQ_QUEUE_FORCE_DELETE = "mq.queue.force.delete"
    MQ_QUEUE_FORCE_DISTRIBUTE = "mq.queue.force.distribute"


from folkmq.client.MqClient import MqClient
from folkmq.client.MqClientDefault import MqClientDefault
from folkmq.client.MqMessage import MqMessage


class FolkMQ:
    @staticmethod
    def version_code()->int:
        return 2

    @staticmethod
    def version_code_as_string()->str:
        return f'{FolkMQ.version_code()}'

    @staticmethod
    def version_name()->str:
        return "1.5.0"

    @staticmethod
    def create_client(*serverUrls) -> MqClient:
        return MqClientDefault(serverUrls)

    @staticmethod
    def new_message(body: str|bytes, key:str|None = None):
        return MqMessage(body,key)





#! /usr/bin/env python
# -*- coding: utf-8 -*_
from setuptools import setup,find_packages

setup(
    name='folkmq',
    version='1.5.0',
    description='@noear/folkmq python project',
    author='noear',
    url='https://folkmq.noear.org/',
    packages=find_packages(exclude=['*folkmq-test*']),   # 包内不需要引用的文件夹
    install_requires=[                          # 依赖包
        'loguru>=0.7.2',
        'websockets>=12.0',
        'socket.d>=2.5.1'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Libraries'
    ],
    zip_safe=True,
    python_requires='>=3.10', # 建议使用3.10及以上
)



import abc

from socketd.transport.core import Entity
from socketd.transport.core.Message import Message
from socketd.transport.core.entity.EntityDefault import EntityDefault

from folkmq.client.MqMessage import MqMessage


class MqMetasResolver:
    #
    # 版本号
    #
    @abc.abstractmethod
    def version(self)-> int:...

    #
    # 获取发送人
    #
    @abc.abstractmethod
    def get_sender(self, m: Entity)-> str:...

    #
    # 获取主建
    #
    @abc.abstractmethod
    def get_key(self, m: Entity)-> str:...

    #
    # 获取标签
    #
    @abc.abstractmethod
    def get_tag(self, m: Entity)-> str:...

    #
    # 获取主题
    #
    @abc.abstractmethod
    def get_topic(self, m: Entity)-> str:...

    #
    # 获取消费者组
    #
    @abc.abstractmethod
    def get_consumer_group(self, m: Entity)-> str:...

    #
    # 设置消费者组
    #
    @abc.abstractmethod
    def set_consumer_group(self, m: Entity, consumerGroup: str):...

    #
    # 获取质量等级（0或1）
    #
    @abc.abstractmethod
    def get_qos(self, m: Entity)-> int:...

    #
    # 获取派发次数
    #
    @abc.abstractmethod
    def get_times(self, m: Entity)-> int:...

    #
    # 设置派发次数
    #
    @abc.abstractmethod
    def set_times(self, m: Entity, times: int):...

    #
    # 获取过期时间
    #
    @abc.abstractmethod
    def get_expiration(self, m: Entity)->float:...

    #
    # 设置过期时间
    #
    @abc.abstractmethod
    def set_expiration(self, m: Entity, expiration: int):...

    #
    # 获取定时时间
    #
    @abc.abstractmethod
    def get_scheduled(self, m: Entity)-> int:...

    #
    # 设置定时时间
    #
    @abc.abstractmethod
    def set_scheduled(self, m: Entity, scheduled: int):...

    #
    # 是否有序
    #
    @abc.abstractmethod
    def is_sequence(self, m: Entity)->bool:...

    #
    # 是否广播
    #
    @abc.abstractmethod
    def is_broadcast(self, m: Entity)->bool:...

    #
    # 是否事务
    #
    @abc.abstractmethod
    def is_transaction(self, m: Entity)->bool:...

    #
    # 设置事务
    #
    @abc.abstractmethod
    def set_transaction(self, m: Entity, isTransaction: bool):...

    #
    # 发布实体构建
    #
    # @param topic   主题
    # @param message 消息
    #
    @abc.abstractmethod
    def publish_entity_build(self, topic: str, message: MqMessage)-> EntityDefault:...

    #
    # 路由消息构建
    #
    # @param topic   主题
    # @param message 消息
    #
    @abc.abstractmethod
    def routing_message_build(self, topic: str, message: MqMessage)-> Message:...